#include <stdlib.h>
#include <string.h>
#include "../plugin/atanua_dll_interface.h"
#include "K8055D_C.h"

static const char plugin_name[]="K8055 Velleman";
static const int chipcount = 1;

/* Get the number of chips this DLL/interface supports */
ATANUADLLEXPORT void ATANUACALL getatanuadllinfo(atanuadllinfo *aDllInfo)
{
    /* DLL interface version. Set to 0. Set to -1 if in error. */
    aDllInfo->mDllVersion = 0;   
    /* Number of chips provided by this DLL */
    aDllInfo->mChipCount = chipcount;    
    /* Pointers to chip names. These may not point to stack! */
    aDllInfo->mChipName = new const char*[chipcount]; 
    /* Category numbers for chips. May be NULL. Categories are 0=dont care, 1=base, 2=chip, 3=I/O */
    aDllInfo->mChipCategory = new int[chipcount]; 

    aDllInfo->mChipName[0] = plugin_name;
    aDllInfo->mChipCategory[0] = 3;
}

static int deviceopen = 0;
static const int pincount = 8+5+4*8;

/* create a chip */
ATANUADLLEXPORT int ATANUACALL create(chipinfostruct *aChipInfo, const char *aChipname)
{
    if (strcmp(aChipname, plugin_name) != 0)
        return 0; // unknown chip

    memset(aChipInfo, 0, sizeof(chipinfostruct));

    if (deviceopen)
        return 0; // already have an instance

    deviceopen = 1;

    /* Number of pins for the structure */
    aChipInfo->mPinCount = pincount;     

    /* Pin output data states. See pinmodes enum. */
    aChipInfo->mPinOutput = new int[pincount];
    int i;
    for (i = 0; i < 8; i++)
        aChipInfo->mPinOutput[i] = PINSTATE_READ;

    for (i = 0; i < 5; i++)
        aChipInfo->mPinOutput[i+8] = PINSTATE_HIGHZ;

    for (i = 0; i < 16; i++)
        aChipInfo->mPinOutput[i+8+5] = PINSTATE_HIGHZ;

    for (i = 0; i < 16; i++)
        aChipInfo->mPinOutput[i+8+5+16] = PINSTATE_READ;

    /* Pin input data states. See netstates enum. */
    aChipInfo->mPinInput = new int[pincount];    

    aChipInfo->mWidth = 13;      /* Chip width */
    aChipInfo->mHeight = 8;     /* Chip height */

    /* Pin coordinates in relation to the chip. Two floats per pin. */
    aChipInfo->mPinCoordinates = new float[pincount*2]; 
    for (i = 0; i < 8; i++)
    {
        // Outputs: right edge
        aChipInfo->mPinCoordinates[i*2+0] = 12.5f;
        aChipInfo->mPinCoordinates[i*2+1] = 3.5f+0.5f * i;
    }
    for (i = 0; i < 5; i++)
    {
        // Inputs: left edge
        aChipInfo->mPinCoordinates[8*2+i*2+0] = 0.0f;
        aChipInfo->mPinCoordinates[8*2+i*2+1] = 3.5f+0.5f * i;
    }
    for (i = 0; i < 2*8; i++)
    {
        // Analog inputs: on the top
        aChipInfo->mPinCoordinates[13*2+i*2+0] = 1.0f + 0.5f * i + ((i > 7)?1.0f:0.0f);
        aChipInfo->mPinCoordinates[13*2+i*2+1] = 0.0f;

        // Analog outputs: at bottom. 
        // Gets mixed up with the USB connector in graphics but what can you do..
        aChipInfo->mPinCoordinates[29*2+i*2+0] = 1.0f + 0.5f * i + ((i > 7)?1.0f:0.0f);
        aChipInfo->mPinCoordinates[29*2+i*2+1] = 7.5f;
    }


    /* Texture filename. Can be NULL. */
    aChipInfo->mTextureFilename = "data/k8055.png"; 

    /* Chip name rendered at the center of the chip */
    aChipInfo->mChipName = "Velleman K8055";        

    /* Chip tooltip. Can be NULL. */
    aChipInfo->mTooltip = "Velleman K8055\nUSB experiment board\n\nDigital I/O sampled at 20Hz\nAnalog I/O sampled at 10Hz";         

    /* Tooltips for the pins. Can be NULL. */
    aChipInfo->mPinTooltips = new const char*[pincount];
    aChipInfo->mPinTooltips[0] = "Out8";
    aChipInfo->mPinTooltips[1] = "Out7";
    aChipInfo->mPinTooltips[2] = "Out6";
    aChipInfo->mPinTooltips[3] = "Out5";
    aChipInfo->mPinTooltips[4] = "Out4";
    aChipInfo->mPinTooltips[5] = "Out3";
    aChipInfo->mPinTooltips[6] = "Out2";
    aChipInfo->mPinTooltips[7] = "Out1";

    aChipInfo->mPinTooltips[8] = "I1";
    aChipInfo->mPinTooltips[9] = "I2";
    aChipInfo->mPinTooltips[10] = "I3";
    aChipInfo->mPinTooltips[11] = "I4";
    aChipInfo->mPinTooltips[12] = "I5";

    aChipInfo->mPinTooltips[13] = "DAC1.1";
    aChipInfo->mPinTooltips[14] = "DAC1.2";
    aChipInfo->mPinTooltips[15] = "DAC1.3";
    aChipInfo->mPinTooltips[16] = "DAC1.4";
    aChipInfo->mPinTooltips[17] = "DAC1.5";
    aChipInfo->mPinTooltips[18] = "DAC1.6";
    aChipInfo->mPinTooltips[19] = "DAC1.7";
    aChipInfo->mPinTooltips[20] = "DAC1.8";

    aChipInfo->mPinTooltips[21] = "DAC2.1";
    aChipInfo->mPinTooltips[22] = "DAC2.2";
    aChipInfo->mPinTooltips[23] = "DAC2.3";
    aChipInfo->mPinTooltips[24] = "DAC2.4";
    aChipInfo->mPinTooltips[25] = "DAC2.5";
    aChipInfo->mPinTooltips[26] = "DAC2.6";
    aChipInfo->mPinTooltips[27] = "DAC2.7";
    aChipInfo->mPinTooltips[28] = "DAC2.8";

    aChipInfo->mPinTooltips[29] = "A1.1";
    aChipInfo->mPinTooltips[30] = "A1.2";
    aChipInfo->mPinTooltips[31] = "A1.3";
    aChipInfo->mPinTooltips[32] = "A1.4";
    aChipInfo->mPinTooltips[33] = "A1.5";
    aChipInfo->mPinTooltips[34] = "A1.6";
    aChipInfo->mPinTooltips[35] = "A1.7";
    aChipInfo->mPinTooltips[36] = "A1.8";

    aChipInfo->mPinTooltips[37] = "A2.1";
    aChipInfo->mPinTooltips[38] = "A2.2";
    aChipInfo->mPinTooltips[39] = "A2.3";
    aChipInfo->mPinTooltips[40] = "A2.4";
    aChipInfo->mPinTooltips[41] = "A2.5";
    aChipInfo->mPinTooltips[42] = "A2.6";
    aChipInfo->mPinTooltips[43] = "A2.7";
    aChipInfo->mPinTooltips[44] = "A2.8";

    /* Number of bytes in persistent data store. Can be 0. */
    aChipInfo->mPersistentDataCount = 0;     

    /* Pointer to the persistent data store. Can be NULL. */
    aChipInfo->mPersistentDataPtr = NULL; 

    /* User pointer for storing other chip-related data. */
    aChipInfo->mUserPtr = NULL;

    OpenDevice(0);

    return 1;
}

/* Update a chip */
ATANUADLLEXPORT void ATANUACALL update(chipinfostruct *aChipInfo, float aTick)
{
    if (!aChipInfo->mPinTooltips)
        return;

    // The K8055 interface is pretty slow, so let's skip some of the 
    // updates and only poll the board at about 20Hz.

    static int tickskip = 0;
    tickskip++;
    if (tickskip < 50) return;
    tickskip = 0;

    int inputs = ReadAllDigital();
    int i;
    for (i = 0; i < 5; i++)
    {
        if (inputs & (1 << i))
        {
            aChipInfo->mPinOutput[i+8] = PINSTATE_WRITE_HIGH;
        }
        else
        {
            aChipInfo->mPinOutput[i+8] = PINSTATE_WRITE_LOW;
        }
    }
    int outputs = 0;
    for (i = 0; i < 8; i++)
        if (aChipInfo->mPinInput[i] != PINSTATE_WRITE_HIGH)
            outputs |= (1 << (7 - i));

    WriteAllDigital(~outputs);

    // Only poll analog I/O at half the rate..
    static int furtherskip = 0;
    furtherskip++;
    if (furtherskip < 2)
        return;
    furtherskip = 0;

    long a1, a2;
    ReadAllAnalog(&a1, &a2);
    for (i = 0; i < 8; i++)
    {
        if (a1 & (1 << i))
        {
            aChipInfo->mPinOutput[i+8+5+8*2] = PINSTATE_WRITE_HIGH;
        }
        else
        {
            aChipInfo->mPinOutput[i+8+5+8*2] = PINSTATE_WRITE_LOW;
        }

        if (a2 & (1 << i))
        {
            aChipInfo->mPinOutput[i+8+5+8*2+8] = PINSTATE_WRITE_HIGH;
        }
        else
        {
            aChipInfo->mPinOutput[i+8+5+8*2+8] = PINSTATE_WRITE_LOW;
        }
    }
    
    a1 = a2 = 0;
    for (i = 0; i < 8; i++)
    {
        if (aChipInfo->mPinInput[i+8+5] != PINSTATE_WRITE_HIGH)
            a1 |= (1 << i);
        if (aChipInfo->mPinInput[i+8+5+8] != PINSTATE_WRITE_HIGH)
            a2 |= (1 << i);
    }
    OutputAllAnalog(a1, a2);

}

/* Render and/or handle key input. Return 0 to let Atanua render the chip. */
ATANUADLLEXPORT int ATANUACALL render(chipinfostruct *aChipInfo, int aKey)
{
    // Let Atanua handle rendering
    return 0;
}

/* Called when chip is being destroyed. */
ATANUADLLEXPORT void ATANUACALL cleanup(chipinfostruct *aChipInfo)
{   
    // Free the bits that we allocated. 
    // Atanua has no idea how we allocated these - could be globals, for instance.
    delete[] aChipInfo->mPinOutput;   
    delete[] aChipInfo->mPinInput;    
    delete[] aChipInfo->mPinCoordinates; 
    delete[] aChipInfo->mPinTooltips;
    if (aChipInfo->mPinTooltips)
    {
        CloseDevice();
        deviceopen = 0;
    }
    return;
}
